use log::debug;
use std::sync::Arc;

use mongodb::bson::{doc, to_bson};

use serenity::{
    async_trait,
    model::{
        application::interaction::application_command::ApplicationCommandInteraction,
        prelude::{
            command::CommandOptionType,
            interaction::{application_command::CommandDataOptionValue, InteractionResponseType},
        },
    },
    prelude::Context,
};

use crate::database::{
    models::{Guild, Tag},
    Client as DatabaseClient,
};

use super::commons::{BotCommand, BotCommandOption, CommandExecutionError};

pub struct CreateTagCommand {
    interaction: ApplicationCommandInteraction,
}

#[async_trait]
impl BotCommand for CreateTagCommand {
    fn name() -> String {
        String::from("create_tag")
    }

    fn description() -> String {
        String::from("Create a new tag")
    }

    fn options_list() -> Vec<BotCommandOption> {
        vec![BotCommandOption {
            name: String::from("tag"),
            description: String::from("The tag to create"),
            kind: CommandOptionType::String,
            required: true,
        }]
    }

    fn new(interaction: ApplicationCommandInteraction) -> Self {
        debug!("Creating a new CreateTagCommand object");
        CreateTagCommand { interaction }
    }

    async fn run(
        &self,
        context: Context,
        database: Arc<DatabaseClient>,
    ) -> Result<(), CommandExecutionError> {
        // Extract tag_name parameter
        let tag_name = match self.interaction.data.options.get(0) {
            Some(a) => match &a.resolved {
                Some(r) => match r {
                    CommandDataOptionValue::String(r_str) => Ok(r_str),
                    _ => Err(CommandExecutionError::ArgumentDeserializationError(
                        "Received non String argument for the CreateTagCommand".to_string(),
                    )),
                },
                None => Err(CommandExecutionError::ArgumentDeserializationError(
                    "Could not deserialize the argument for the CreateTagCommand".to_string(),
                )),
            },
            None => Err(CommandExecutionError::ArgumentExtractionError(
                "Failed to get the CreateTagCommand argument".to_string(),
            )),
        }?;

        // Extract guild id from Serenity context
        let guild_id = match self.interaction.guild_id {
            Some(a) => Ok(a.to_string()),
            None => Err(CommandExecutionError::ContextRetrievalError(
                "Could not fetch guild id from issued command".to_string(),
            )),
        }?;

        let guild = match database.get_by_id::<Guild>(&guild_id).await {
            Ok(query) => match query {
                Some(r) => Ok(r),
                None => Err(CommandExecutionError::ContextRetrievalError(
                    "Failed to retrieve the guild where the command was issued".to_string(),
                )),
            },
            Err(()) => Err(CommandExecutionError::DatabaseQueryError(
                "Could not access to the database".to_string(),
            )),
        }?;

        let matching_tag = guild.tags.iter().find(|t| t.name == *tag_name);

        let response_content = if matching_tag.is_some() {
            String::from("This tag already exist for this server.")
        } else {
            let mut new_tags = guild.tags.clone();
            new_tags.push(Tag {
                name: tag_name.to_string(),
                is_nsfw: false,
                subscribers: vec![],
            });

            match database
                .update_one(
                    guild,
                    doc! {"$set": { "tags": to_bson(&new_tags).unwrap() }},
                )
                .await
            {
                Ok(_) => Ok(String::from("Tag successfully created.")),
                Err(_) => Err(CommandExecutionError::DatabaseQueryError(
                    "Could not add new tag to the database".to_string(),
                )),
            }?
        };

        match self
            .interaction
            .create_interaction_response(context.http, |response| {
                response
                    .kind(InteractionResponseType::ChannelMessageWithSource)
                    .interaction_response_data(|message| message.content(response_content))
            })
            .await
        {
            Ok(_) => Ok(()),
            Err(_e) => Err(CommandExecutionError::DiscordAPICallError(
                "Failed to answer to the initial command".to_string(),
            )),
        }
    }
}
